#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch

__all__ = ["gradient_with_pbc",
           "first_derivative_with_pbc"]

def _get_diff_slices(ndim, xdim):
    # ndim ... number of dimensions
    # xdim ... dimension that is used for differentiation
    slice_0 = [slice(None)] * ndim # for a nx,ny,nz,nd tensor, this is eqivalent to :,:,:,:
    slice_1 = [slice(None)] * ndim
    slice_0[xdim] = slice(0, -1)
    slice_1[xdim] = slice(1, None)
    slice_0 = tuple(slice_0)
    slice_1 = tuple(slice_1)

    return slice_0, slice_1

def _get_derivative_on_boundary(f0,f1,f2,h0,h1):
    g_bdr = -f2*h0**2. + f1*(h0+h1)**2. - f0*h1*(2*h0+h1)
    g_bdr /= h0*h1*(h0 + h1)
    return g_bdr

def gradient_with_pbc(f, mesh, dim=[0,1,2], C=None, Bl=None, Br=None, slices=None, second_order_boundary=False):
    """
    f      ... the gradient of this tensor is calculated 
    C,B    ... assumed to be constants on cells, gives jump conditions of the type
            lim x->a C_0 * df(x)/dx + B0 = lim a<-x C_1 * df(x)/dx + B1 
    mesh   ... the mesh on which f lives
    dim    ... the dimensions in which the first derivatives are calculated
    slices ... slicing of f, default to :,:,:
    """
    # validate input for dim
    dims = [dim] if isinstance(dim, int) else dim
    assert isinstance(dims, list) and all(isinstance(d, int) for d in dims)

    # slices 
    if slices == None:
        slices = tuple([slice(None)]*3)
    assert isinstance(slices, tuple) and all(isinstance(s, slice) for s in slices)

    # get jump condition data
    if (Bl is None) == (Br is None):
        if Bl is None:
            Bl = [torch.zeros(mesh.n)]*len(dims)
            Br = [torch.zeros(mesh.n)]*len(dims)
        Bl = Bl if isinstance(Bl, list) else [Bl]
        Br = Br if isinstance(Br, list) else [Br]
        assert len(Bl)==len(dims) and len(Br)==len(dims)

    else:
        raise Exception("If Bl is set, Br needs to be set too and vice-versa.")
    
    if C is None:
        C = [torch.ones(mesh.n)]*len(dims)
    C = C if isinstance(C, list) else [C]
    assert len(C)==len(dims)

    # slice data 
    fs = f[slices]
    n_f = fs.shape # used nx, ny, nz + data point dimensions
    
    # get expanded dx and slice it
    dx = mesh.dx_tensor[0][slices[0]].reshape(-1,1,1)
    dy = mesh.dx_tensor[1][slices[1]].reshape(1,-1,1)
    dz = mesh.dx_tensor[2][slices[2]].reshape(1,1,-1)
    dx_exp = dx,dy,dz

    # get gradient
    output = []
    for i in range(len(dims)):
        d = dims[i]
        # if there is only one data point in this dimension, the first derivative is 0
        if (n_f[d]) < 2:
            output.append(torch.zeros(n_f))
        # case no pbc: boundary node order is 1, obtained from forward and backward difference
        # this method is also used when the slice limits the input to within the mesh
        elif ((mesh.pbc[d] == 0) or (fs.shape[d] != f.shape[d])):
            g = first_derivative_with_jump_conditions(fs, C[i][slices], Bl[i][slices], Br[i][slices], dx_exp[d], d)
            
            if (n_f[d] > 2) and second_order_boundary:
                dx_s = dx_exp[d]
                if (d == 0):
                    f0 = fs[0]
                    f1 = fs[1]
                    f2 = fs[2]
                    h0 = 0.5*(dx_s[0]+dx_s[1])
                    h1 = 0.5*(dx_s[1]+dx_s[2])
                    g[0] = _get_derivative_on_boundary(f0, f1, f2, h0, h1)

                    f0 = fs[-1]
                    f1 = fs[-2]
                    f2 = fs[-3]
                    h0 = 0.5*(dx_s[-1]+dx_s[-2])
                    h1 = 0.5*(dx_s[-2]+dx_s[-3])
                    g[-1] = -_get_derivative_on_boundary(f0, f1, f2, h0, h1)

                if (d == 1):
                    f0 = fs[:,0]
                    f1 = fs[:,1]
                    f2 = fs[:,2]
                    h0 = 0.5*(dx_s[:,0]+dx_s[:,1])
                    h1 = 0.5*(dx_s[:,1]+dx_s[:,2])
                    g[:,0] = _get_derivative_on_boundary(f0, f1, f2, h0, h1)

                    f0 = fs[:,-1]
                    f1 = fs[:,-2]
                    f2 = fs[:,-3]
                    h0 = 0.5*(dx_s[:,-1]+dx_s[:,-2])
                    h1 = 0.5*(dx_s[:,-2]+dx_s[:,-3])
                    g[:,-1] = -_get_derivative_on_boundary(f0, f1, f2, h0, h1)

                if (d == 2):
                    f0 = fs[:,:,0]
                    f1 = fs[:,:,1]
                    f2 = fs[:,:,2]
                    h0 = 0.5*(dx_s[:,:,0]+dx_s[:,:,1])
                    h1 = 0.5*(dx_s[:,:,1]+dx_s[:,:,2])
                    g[:,:,0] = _get_derivative_on_boundary(f0, f1, f2, h0, h1)

                    f0 = fs[:,:,-1]
                    f1 = fs[:,:,-2]
                    f2 = fs[:,:,-3]
                    h0 = 0.5*(dx_s[:,:,-1]+dx_s[:,:,-2])
                    h1 = 0.5*(dx_s[:,:,-2]+dx_s[:,:,-3])
                    g[:,:,-1] = -_get_derivative_on_boundary(f0, f1, f2, h0, h1)

            output.append(g)
        # case pbc
        else:
            g = first_derivative_with_jump_conditions_and_pbc(fs, C[i][slices], Bl[i][slices], Br[i][slices], dx_exp[d], d)
            output.append(g)

    return output

def first_derivative_with_jump_conditions(f, C, Bl, Br, dx, dim):   
    slice_0, slice_1 = _get_diff_slices(f.ndim, dim)

    # for non-periodic meshes: 
    # the virtual node outside the boundary is assumed to have the same dx as the boundary node

    h = 0.5*(dx[slice_0] + dx[slice_1])
    h_sum = torch.zeros(dx.shape)
    h_sum[slice_0] = h 
    h_sum[slice_1] += h

    # other aliases
    C0 = C[slice_0]
    C1 = C[slice_1]

    B0 = Br[slice_0]
    B1 = Bl[slice_1]

    f0 = f[slice_0]
    f1 = f[slice_1]

    # output gradient component should be equivalent to torch.gradient if C and B are constant
    g = torch.zeros(f.shape)
    
    denom = (C0*dx[slice_1]+C1*dx[slice_0])

    # backward difference
    g[slice_1] += h*(2.*C0*(f1-f0) + dx[slice_0]*(B0-B1)) / denom

    # forward difference
    g[slice_0] += h*(2.*C1*(f1-f0) + dx[slice_1]*(B1-B0)) / denom

    # weighted mean between forward and backward difference 
    return g.nan_to_num(posinf=0, neginf=0)/h_sum

def first_derivative_with_jump_conditions_and_pbc(f, C, Bl, Br, dx, dim):
    fl = torch.roll(f, 1, dims=dim)
    fr = torch.roll(f, -1, dims=dim)

    Cl = torch.roll(C, 1, dims=dim)
    Cr = torch.roll(C, -1, dims=dim)

    dxl = torch.roll(dx, 1, dims=dim)
    dxr = torch.roll(dx, -1, dims=dim)

    # spacing between the data points
    hl = 0.5*(dx+dxl)
    hr = 0.5*(dx+dxr)

    # first-order accurate approximations of first derivative
    g_bkw = (2.*Cl*(f-fl) + dxl*(torch.roll(Br,  1, dims=dim)-Bl)) / (Cl*dx+C*dxl)
    g_fwd = (2.*Cr*(fr-f) + dxr*(torch.roll(Bl, -1, dims=dim)-Br)) / (C*dxr+Cr*dx)
    
    g_bkw = g_bkw.nan_to_num(posinf=0, neginf=0)
    g_fwd = g_fwd.nan_to_num(posinf=0, neginf=0)

    # weighted mean between forward and backward difference 
    # this is second-order accurate, when C and B are constant
    return (hr*g_bkw + hl*g_fwd)/(hl+hr)

@torch.compile
def first_derivative_with_pbc(f, dx_expanded, dim):
    fl = torch.roll(f, 1, dims=dim)
    fr = torch.roll(f, -1, dims=dim)

    hl = 0.5*(dx_expanded + torch.roll(dx_expanded, 1, dims=dim))
    hr = torch.roll(hl, -1, dims=dim)

    return  (hl**2.*fr - hr**2.*fl + (hr**2.-hl**2.)*f) / (hr*hl**2. + hl*hr**2)


#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import logging, timedmethod, constants, normalize
from . import LLGSolver
from magnumnp.linear_elasticity.bcs import Plane, PlaneBC
from magnumnp.linear_elasticity.deriv_term_compiler import *
from magnumnp.linear_elasticity.strain import epsilon, epsilon_m, _get_C_jump_conditions, _get_B_jump_conditions, _get_sigM_jump_conditions
from magnumnp.linear_elasticity.stress import sigma
from magnumnp.linear_elasticity.utils import gradient_with_pbc, _get_diff_slices

from .ode_solvers import RKF45
import torch
import numpy as np

__all__ = ["LLGWithLESolver"]

class LLGWithLESolver(LLGSolver):
    r"""
    Extension of the ``LLGSolver`` class that also yields the rate of change of 
    mechanical displacement and momentum density.

    When used with the magnetoelastic field, this solver enables self-consistent 
    time integration of magnetization dynamics and elastodynamics.

    Examples
    --------
    Standard usage:

        .. code:: python

            llg = LLGWithLESolver([magnetoelastic, demag, exchange, external])
            logger = Logger("data", fields=['m', 'ud', 'pd'])
            while state.t < 1e-9 - eps:
                llg.step(state, 1e-11)
                logger << state

    Optimizations for speed and lower memory usage:

    1. Cubic material with a magnetic film in the 3rd x-layer:

        .. code:: python

            llg = LLGWithLESolver(
                [magnetoelastic, anisotropy, demag, exchange, external],
                C_sym="cubic",
                magnetic_x_limits=[3, 4]
            )

    2. Material with a magnetic domain on [:, 2:, :4]:

        .. code:: python

            llg = LLGWithLESolver(
                [magnetoelastic, demag, exchange, external],
                magnetic_y_limits=[3, None],
                magnetic_z_limits=[0, 4],
                boundary_nodes=2
            )

    3. 1D setup with pbcs:

        .. code:: python

            llg = LLGWithLESolver(
                [magnetoelastic, anisotropy, exchange, external],
                iteration_depth = 0
            )

    Parameters
    ----------
    terms : list of :class:`LLGTerm`
        List of LLG contributions to be considered for time integration.
    solver : :class:`Solver`, optional
        ODE solver to use. Choose one of: ``RKF45`` (default), ``RKF56``, 
        ``ScipyODE``, ``ScipyOdeint``, ``TorchDiffEq``, or ``TorchDiffEqAdjoint``.
    no_precession : bool, optional
        If ``True``, integrates without the precession term (default: ``False``).
    mask_elastic : torch.Tensor, optional
        Boolean or integer mask of shape ``mesh.n`` restricting the region where 
        elastodynamic properties are updated (default: ``None``).
    C_sym : numpy.ndarray or str, optional
        Boolean or integer mask of shape ``(6, 6)`` restricting which stiffness 
        tensor elements are considered (default: ``None``). Can also be a string 
        (``"isotropic"`` or ``"cubic"``) to include only C11, C12, C13, C22, C23, 
        C33, C44, C55, and C66.
    magnetic_x_limits : list of int, optional
        Two-element list giving start and stop indices for slicing the x-direction 
        of the magnetic domain (default: ``None``).
    magnetic_y_limits : list of int, optional
        Two-element list giving start and stop indices for slicing the y-direction 
        of the magnetic domain (default: ``None``).
    magnetic_z_limits : list of int, optional
        Two-element list giving start and stop indices for slicing the z-direction 
        of the magnetic domain (default: ``None``).
    boundary_nodes : int, optional
        Number of boundary nodes for boundary value computations (default: ``1``). 
        Can be 1, 2, or 3. In setups with multiple material domains, this value 
        cannot exceed the number of boundary cells in the same domain, counting 
        from the interface for each interface individually.
    iteration_depth : int, optional
        Iteration depth for the strain jump conditions (default: ``1``).
    """

    def __init__(self, 
                 terms,                     # magnetic field terms
                 solver = RKF45,            # used solver
                 no_precession = False,     # remove the precessional term from the LLG
                 mask_elastic = None,       # torch.tensor (bool), indicates where ud and pd are updated
                 C_sym = None,              # 2D list, 6 times 6, entries of 1 indicate that the corresponding stiffness tensor component is non-zero. 0 indicates that it is zero.
                 magnetic_x_limits = None,  # list of len 2, upper and lower limit of the magnetic domain in x-direction
                 magnetic_y_limits = None,  # list of len 2, upper and lower limit of the magnetic domain in y-direction
                 magnetic_z_limits = None,  # list of len 2, upper and lower limit of the magnetic domain in z-direction
                 boundary_nodes = 1,        # number of boundary nodes used for boundary value computations
                 iteration_depth = 1,       # iteration depth for strain jump conditions
                 **kwargs):

        # field terms for llg time integration
        self._terms = terms
        self._no_precession = no_precession

        # ode solver
        self._solver = solver(self.dv, **kwargs)
         
        # mask for state.pd and state.ud update
        self._additional_mask_elastic = mask_elastic

        # limit of the magnetic domain
        if magnetic_x_limits == None:
            self._magnetic_x_limits = [0, None]
        else:
            self._magnetic_x_limits = magnetic_x_limits
        if magnetic_y_limits == None:
            self._magnetic_y_limits = [0, None]
        else:
            self._magnetic_y_limits = magnetic_y_limits
        if magnetic_z_limits == None:
            self._magnetic_z_limits = [0, None]
        else:
            self._magnetic_z_limits = magnetic_z_limits

        ix0, ix1 = self._magnetic_x_limits
        iy0, iy1 = self._magnetic_y_limits
        iz0, iz1 = self._magnetic_z_limits

        self.slice_m = tuple([slice(ix0, ix1), slice(iy0, iy1), slice(iz0, iz1)])

        # mask for the stiffness matrix
        if isinstance(C_sym, str):
            if C_sym == "cubic" or C_sym == "isotropic":
                self._C_mask = [[1, 1, 1, 0, 0, 0],
                                [1, 1, 1, 0, 0, 0],
                                [1, 1, 1, 0, 0, 0],
                                [0, 0, 0, 1, 0, 0],
                                [0, 0, 0, 0, 1, 0],
                                [0, 0, 0, 0, 0, 1]]
            else:
                raise Exception(self.__class__.__name__ + ": C_sym '" + C_sym + "' is unknown. C_sym can be 'isotropic' 'cubic' or a mask of shape (6,6)")
        elif C_sym is None:
            self._C_mask = []
            self._C_mask.append(6*[1])
            for i in range(5):
                self._C_mask.append(self._C_mask[-1])
        else:
            if isinstance(C_sym, torch.Tensor):
                C_sym = C_sym.detach().cpu().numpy()
            if isinstance(C_sym, (list, np.ndarray)):
                C_sym = np.array(C_sym)
            if not isinstance(C_sym, np.ndarray) or C_sym.shape != (6, 6):
                raise Exception(self.__class__.__name__+": C_sym was not successfully converted into a (6,6) array")
            self._C_mask = C_sym

        self.iteration_depth = iteration_depth

        if (boundary_nodes < 1) or (boundary_nodes > 3):
            raise Exception(self.__class__.__name__ + ": boundary_nodes must be 1, 2 or 3.")
        self._gradient_second_order_boundary = boundary_nodes > 2 
        self._bc_second_order = boundary_nodes > 1

        # neumann bcs
        self._neumann_bcs = []

        # obtain the individual terms that make up the stress matrix and the force components
        # the former is organized as a 3-dimensional list:
        #       the first two dimensions being the matrix components of the stress matrix
        #       the last dimension contains the individual terms making up the corresponding stress component (lambda functions)
        # the latter is organized as follows: 
        #       the first dimension contains the components of the force, i.e. 0 = x, 1 = y, 2 = z,
        #       such that terms that contribute to fx are found in self._f_terms[0]
        #       the second dimension separates the components by their second derivative, 
        #       i.e. 0 = terms with second derivative in x-direction (e.g. Dx[C[0,1]*Dy[u[1]]]), ...
        #       the last dimension contains the individual terms (lambda functions)

        self._sig_terms, self._f_terms, self._fm_terms = self._compile_terms()

    # -----------------------------------
    # Definition of the solution variable
    # -----------------------------------

    #@torch.compile  
    def _get_solution_variables(self, state):
        n = state.mesh.n
        v = torch.zeros((n+(9,)))

        v[:,:,:,:3] = state.m 
        v[:,:,:,6:] = state.pd

        ud = state.ud
        if hasattr(state, "bcs"):  
            if "ud" in state.bcs:
                for bc in state.bcs["ud"]:
                    ud[bc.mask] = torch.flatten(bc.condition(state), end_dim=-2)

        v[:,:,:,3:6] = ud

        #v.detach().cpu().numpy().reshape(-1, order = 'F')

        return v
    
    #@torch.compile  
    def _set_solution_variables(self, state, v):
        # v = state.Tensor(v.reshape(state.mesh.n + (9,), order = "F"))

        state.m = v[:,:,:,:3]
        state.ud = v[:,:,:,3:6]
        state.pd = v[:,:,:,6:]

        if hasattr(state, "bcs"):
            if "ud" in state.bcs:
                for bc in state.bcs["ud"]:
                    state.ud[bc.mask] = torch.flatten(bc.condition(state), end_dim=-2)
    
    # ------------------------------
    # Definition of the problem area
    # ------------------------------

    def _get_mask_elastic(self, state):
        mask = self._additional_mask_elastic
        n = state.mesh.n

        if mask == None:
            mask = torch.ones(n, dtype=torch.bool)

        if hasattr(state, "bcs"):
            if "ud" in state.bcs:
                for bc in state.bcs["ud"]:
                    mask = torch.logical_and(mask, torch.logical_not(bc.mask))
        
        elastic_mask = torch.zeros(n+(3,), dtype=torch.bool)
        elastic_mask[:,:,:,0] = mask
        elastic_mask[:,:,:,1] = mask
        elastic_mask[:,:,:,2] = mask

        return elastic_mask
    
    # ----------------------------------------------------
    # Construction of the PDE RHS and stress (sigma) terms
    # ----------------------------------------------------
    
    def _compile_terms(self):
        compiler = DerivTermCompiler()

        # -------------
        # Collect terms
        # -------------

        # Initialize all strain terms
        # ... in Voigt notation
        eps = []
        eps.append([EpsTerm(i_u=0, i_x=0)])
        eps.append([EpsTerm(1,1)])
        eps.append([EpsTerm(2,2)])
        eps.append([EpsTerm(1,2), EpsTerm(2,1)])
        eps.append([EpsTerm(0,2), EpsTerm(2,0)])
        eps.append([EpsTerm(0,1), EpsTerm(1,0)])

        eps_m = []
        eps_m.append(EpsMTerm(0,0))
        eps_m.append(EpsMTerm(1,1))
        eps_m.append(EpsMTerm(2,2))
        eps_m.append(EpsMTerm(1,2))
        eps_m.append(EpsMTerm(0,2))
        eps_m.append(EpsMTerm(0,1))

        # multiply them by the stiffness tensor and obtain the stress terms
        sig_v = []
        sig_m_v = []
        for i in range(6):
            terms = []
            terms_m = []
            for j in range(6):
                # if the user informed the solver about the symmetry of the problem, exclude terms with a zero-valued stiffness tensor component
                if self._C_mask[i][j] == 0:
                    logging.info_green("Warning: due to the selected symmetry of C, a term with C_{"+str(i)+","+str(j)+"} was excluded!")
                else:
                    terms_m.append(eps_m[j].multiply_Cij(i,j))
                    for eps_term in eps[j]:
                        terms.append(eps_term.multiply_Cij(i,j))

            sig_v.append(terms)
            sig_m_v.append(terms_m)

        # rearrange the stress terms from Voigt notation into matrix form
        sig = []
        sig.append([sig_v[0], sig_v[5], sig_v[4]])
        sig.append([sig_v[5], sig_v[1], sig_v[3]])
        sig.append([sig_v[4], sig_v[3], sig_v[2]])

        sig_m = []
        sig_m.append([sig_m_v[0], sig_m_v[5], sig_m_v[4]])
        sig_m.append([sig_m_v[5], sig_m_v[1], sig_m_v[3]])
        sig_m.append([sig_m_v[4], sig_m_v[3], sig_m_v[2]])

        # -------------
        # compile terms
        # -------------

        sig_lambdas = []
        f_lambdas = []
        fm_lambdas = []
        for i in range(3):
            sig_row = []
            f_row = []
            fm_row = []
            for j in range(3):
                sig_element = []
                f_element = []
                fm_element = []
                for sig_term in sig[i][j]:
                    term_function = compiler.compile(self, sig_term)
                    sig_element.append(term_function)

                    term = sig_term.differentiate(j)
                    term_function = compiler.compile(self, term)
                    f_element.append(term_function)

                for sig_m_term in sig_m[i][j]:
                    term = sig_m_term.differentiate(j)
                    term_function = compiler.compile(self, term)
                    fm_element.append(term_function)
                
                sig_row.append(sig_element)
                f_row.append(f_element)
                fm_row.append(fm_element)

            sig_lambdas.append(sig_row)
            f_lambdas.append(f_row)
            fm_lambdas.append(fm_row)

        return sig_lambdas, f_lambdas, fm_lambdas

    # ----------------------------------------------------
    # Definition of 1st, 2nd, and mixed derivatives
    # ----------------------------------------------------

    class DiffData:
        # Object that holds forward differences and first derivatives obtained from the midpoint rule
        def __init__(self):
            self._gradient_ud = []
            self._gradient_m = []

        def set_expanded_dx(self, state):
            dx_exp, dy_exp, dz_exp = state.mesh.dx_tensor

            dx_exp = dx_exp.unsqueeze(1).unsqueeze(2)
            dy_exp = dy_exp.unsqueeze(0).unsqueeze(2)
            dz_exp = dz_exp.unsqueeze(0).unsqueeze(0)

            dx_exp = dx_exp.expand_as(torch.zeros(state.mesh.n))
            dy_exp = dy_exp.expand_as(torch.zeros(state.mesh.n))
            dz_exp = dz_exp.expand_as(torch.zeros(state.mesh.n))
            self.dx_epx = dx_exp, dy_exp, dz_exp

        def set_gradient_ud(self, grad_ux, grad_uy, grad_uz):
            data = []
            data.append(grad_ux)
            data.append(grad_uy)
            data.append(grad_uz)

            self._gradient_ud = data

        def set_gradient_m(self, grad_mx, grad_my, grad_mz):
            data = []
            data.append(grad_mx)
            data.append(grad_my)
            data.append(grad_mz)

            self._gradient_m = data

        def set_B_jump_conditions(self, Bl, Br):
            self._Bl_jump_conditions = Bl
            self._Br_jump_conditions = Br

        @property
        def gradient_ud(self):
            return self._gradient_ud
        
        @property
        def gradient_m(self):
            return self._gradient_m
    
    def _2nd_derivative(self, state, i_u, i_x, ij_C):
        if (state.mesh.pbc[i_x] == 0) and (state.mesh.n[i_x]>1):
            return self._2nd_derivative_homogeneous_Neumann(state, i_u, i_x, ij_C)
        else:
            return self._2nd_derivative_with_pbc(state, i_u, i_x, ij_C)

    #@torch.compile
    def _2nd_derivative_homogeneous_Neumann(self, state, i_u, i_x, ij_C):
        a = torch.zeros(state.mesh.n)

        dx = self.diff_data.dx_epx[i_x]

        slice_0, slice_1 = _get_diff_slices(3, i_x)

        C = state.material["C"][:,:,:,ij_C[0],ij_C[1]]

        C_denom = (dx[slice_0]*C[slice_1] + dx[slice_1]*C[slice_0])
        C_avg = 2.*C[slice_0]*C[slice_1] / C_denom
        C_avg.nan_to_num(posinf=0, neginf=0) # C could be 0 if not proper C_mask is set

        # Look up: [Bx_xl, Bx_yl, Bx_zl], [By_xl, By_yl, By_zl], [Bz_xl, Bz_yl, Bz_zl]
        Bl = self.diff_data._Bl_jump_conditions[i_u][i_x]
        Br = self.diff_data._Br_jump_conditions[i_u][i_x]
        jump = -(Br[slice_0] - Bl[slice_1]) / C_denom
        jump.nan_to_num(posinf=0, neginf=0)

        diff = state.ud[slice_1+(i_u,)] - state.ud[slice_0+(i_u,)] 

        a[slice_0] += C_avg*diff # (ud_i+1 - ud_i) * (C_i+1*dx_i + dx_1+i*C_i) / (2*C_i+1*C_i)
        a[slice_1] -= C_avg*diff # -(ud_i - ud_i-1) * (dx_i-1*C_i + dx_i*C_i-1)

        a[slice_0] += dx[slice_1]*C[slice_0]*jump
        a[slice_1] += dx[slice_0]*C[slice_1]*jump

        set_slices = [slice(None)]*dx.ndim 
        set_slices[i_x] = slice(1, -1)
        set_slices = tuple(set_slices)

        l_slices = [slice(None)]*dx.ndim 
        l_slices[i_x] = slice(2,None)
        l_slices = tuple(l_slices)

        r_slices = [slice(None)]*dx.ndim 
        r_slices[i_x] = slice(None,-2)
        r_slices = tuple(r_slices)

        dx_denom = torch.clone(dx)
        dx_denom[set_slices] += 0.5 * (dx[l_slices]+dx[r_slices])
        dx_denom[set_slices] /= 2.
        
        return (a/(dx_denom)).nan_to_num(posinf=0, neginf=0)
    
    #@torch.compile
    def _2nd_derivative_with_pbc(self, state, i_u, i_x, ij_C):
        # second order accurate for regular and only first order accurate for irregular grids
        a = torch.zeros(state.mesh.n)

        dx = self.diff_data.dx_epx[i_x]
        C = state.material["C"][:,:,:,ij_C[0],ij_C[1]]

        C_next = torch.roll(C, -1, dims=i_x) # positive shift, to align this with the definition of the forward differences
        dx_next = torch.roll(dx, -1, dims=i_x) # positive shift, to align this with the definition of the forward differences
        C_denom = (C_next*dx + C*dx_next) # at 0: C_1*dx_0 + C_0*dx_1, at N: C_0*dx_N + C_N*dx_0
        C_avg = 2.*C_next * C / C_denom
        C_avg.nan_to_num(posinf=0, neginf=0) # C could be 0 if not proper C_mask is set

        diff = torch.roll(state.ud[...,i_u],-1,i_x) - state.ud[...,i_u]
        a += C_avg * diff
        a -= torch.roll(a, +1, i_x)

        # Look up: [Bx_xl, Bx_yl, Bx_zl], [By_xl, By_yl, By_zl], [Bz_xl, Bz_yl, Bz_zl]
        Bl = self.diff_data._Bl_jump_conditions[i_u][i_x]
        Br = self.diff_data._Br_jump_conditions[i_u][i_x]

        jump = -(Br - torch.roll(Bl, -1, i_x)) / C_denom
        jump.nan_to_num(posinf=0, neginf=0) # C could be 0 if not proper C_mask is set
        
        a += torch.roll(dx, -1, i_x)*C*jump 
        a += torch.roll(dx, +1, i_x)*C*torch.roll(jump, +1, i_x)
        
        dx_denom = torch.clone(dx)
        dx_denom += 0.5*(torch.roll(dx, +1, i_x) + torch.roll(dx, -1, i_x))
        dx_denom /= 2.

        return (a/dx_denom).nan_to_num(posinf=0, neginf=0)
    
    #@torch.compile
    def _mixed_derivative(self, state, i_u, i_x1, i_x2, ij_C):
        diff_1st = self.diff_data.gradient_ud[i_u][i_x1]
        C = state.material["C"][:,:,:,ij_C[0],ij_C[1]]

        # Note: using C jump conditions here allows to handle jumps to vaccum correctly
        a = gradient_with_pbc(C*diff_1st, state.mesh, dim=i_x2, C=[C], second_order_boundary=self._gradient_second_order_boundary)[0]

        return a
    
    #@torch.compile
    def _weighted_first_derivative(self, state, i_u, i_x, ij_C):
        C = state.material["C"][:,:,:,ij_C[0],ij_C[1]]
        return C * self.diff_data.gradient_ud[i_u][i_x]
    
    #@torch.compile 
    def _main_diag_sigM_derivative(self, state, i_m, i_x, ij_C):
        slice_m = self.slice_m
        l100 = state.material["lambda_100"][slice_m+(0,)]
        C = state.material["C"][slice_m+(ij_C[0],ij_C[1])]
        mi = state.m[slice_m+(i_m,)]
        dmi = self.diff_data.gradient_m[i_m][i_x]

        return 3.*l100*C*mi*dmi 
    
    #@torch.compile
    def _off_diag_sigM_derivative(self, state, i_m, j_m, i_x, ij_C):
        slice_m = self.slice_m
        l111 = state.material["lambda_111"][slice_m+(0,)]
        C = state.material["C"][slice_m+(ij_C[0],ij_C[1])]
        mi = state.m[slice_m+(i_m,)]
        mj = state.m[slice_m+(j_m,)]
        dmi = self.diff_data.gradient_m[i_m][i_x]
        dmj = self.diff_data.gradient_m[j_m][i_x]

        return 3.*l111*C*(mi*dmj + mj*dmi)

    # ----------------------------------------------------
    # Helpers
    # ----------------------------------------------------

    def _update_diff_data(self, state):
        diff_data = self.DiffData()
        diff_data.set_expanded_dx(state)

        # Get 2nd order 1st derivatives of m
        A = state.material["A"][...,0]

        slc_m = self.slice_m
        grad_mx = gradient_with_pbc(state.m[...,0], state.mesh, [0,1,2], [A, A, A], slices=slc_m, second_order_boundary=self._gradient_second_order_boundary)
        grad_my = gradient_with_pbc(state.m[...,1], state.mesh, [0,1,2], [A, A, A], slices=slc_m, second_order_boundary=self._gradient_second_order_boundary)
        grad_mz = gradient_with_pbc(state.m[...,2], state.mesh, [0,1,2], [A, A, A], slices=slc_m, second_order_boundary=self._gradient_second_order_boundary)

        diff_data.set_gradient_m(grad_mx, grad_my, grad_mz)

        # Get 2nd order 1st derivatives of u 
        mxl = torch.clone(state.m)
        mxr = torch.clone(state.m)
        myl = torch.clone(state.m)
        myr = torch.clone(state.m)
        mzl = torch.clone(state.m)
        mzr = torch.clone(state.m)

        dx_exp = state.mesh.dx_tensor[0].reshape(-1,1,1,1)
        dy_exp = state.mesh.dx_tensor[1].reshape(1,-1,1,1)
        dz_exp = state.mesh.dx_tensor[2].reshape(1,1,-1,1)

        dmdx = torch.zeros(state.mesh.n+(3,))
        dmdx[slc_m+(0,)] = grad_mx[0]
        dmdx[slc_m+(1,)] = grad_my[0]
        dmdx[slc_m+(2,)] = grad_mz[0]

        dmdy = torch.zeros(state.mesh.n+(3,))
        dmdy[slc_m+(0,)] = grad_mx[1]
        dmdy[slc_m+(1,)] = grad_my[1]
        dmdy[slc_m+(2,)] = grad_mz[1]

        dmdz = torch.zeros(state.mesh.n+(3,))
        dmdz[slc_m+(0,)] = grad_mx[2]
        dmdz[slc_m+(1,)] = grad_my[2]
        dmdz[slc_m+(2,)] = grad_mz[2]

        mxl -= 0.5*dmdx*dx_exp
        mxr += 0.5*dmdx*dx_exp
        myl -= 0.5*dmdy*dy_exp
        myr += 0.5*dmdy*dy_exp
        mzl -= 0.5*dmdz*dz_exp
        mzr += 0.5*dmdz*dz_exp

        m_data = [mxl, mxr], [myl, myr], [mzl, mzr]
        C = _get_C_jump_conditions(state)
        Bl_sigM, Br_sigM = _get_sigM_jump_conditions(state, m_data)

        grad_ud_x = gradient_with_pbc(state.ud[...,0], state.mesh, [0,1,2], C[0], Bl_sigM[0], Br_sigM[0], second_order_boundary=self._gradient_second_order_boundary)
        grad_ud_y = gradient_with_pbc(state.ud[...,1], state.mesh, [0,1,2], C[1], Bl_sigM[1], Br_sigM[1], second_order_boundary=self._gradient_second_order_boundary)
        grad_ud_z = gradient_with_pbc(state.ud[...,2], state.mesh, [0,1,2], C[2], Bl_sigM[2], Br_sigM[2], second_order_boundary=self._gradient_second_order_boundary)

        diff_data.set_B_jump_conditions(Bl_sigM, Br_sigM)

        if (self.iteration_depth > 0):
            for iter in range(self.iteration_depth):
                gradient_data = grad_ud_x, grad_ud_y, grad_ud_z

                Bl_eps, Br_eps = _get_B_jump_conditions(state, gradient_data)

                Bxl = [Bl_eps[0][0] + Bl_sigM[0][0], Bl_eps[0][1] + Bl_sigM[0][1], Bl_eps[0][2] + Bl_sigM[0][2]]
                Byl = [Bl_eps[1][0] + Bl_sigM[1][0], Bl_eps[1][1] + Bl_sigM[1][1], Bl_eps[1][2] + Bl_sigM[1][2]]
                Bzl = [Bl_eps[2][0] + Bl_sigM[2][0], Bl_eps[2][1] + Bl_sigM[2][1], Bl_eps[2][2] + Bl_sigM[2][2]]

                Bxr = [Br_eps[0][0] + Br_sigM[0][0], Br_eps[0][1] + Br_sigM[0][1], Br_eps[0][2] + Br_sigM[0][2]]
                Byr = [Br_eps[1][0] + Br_sigM[1][0], Br_eps[1][1] + Br_sigM[1][1], Br_eps[1][2] + Br_sigM[1][2]]
                Bzr = [Br_eps[2][0] + Br_sigM[2][0], Br_eps[2][1] + Br_sigM[2][1], Br_eps[2][2] + Br_sigM[2][2]]

                Bl = Bxl, Byl, Bzl 
                Br = Bxr, Byr, Bzr 

                grad_ud_x[:] = gradient_with_pbc(state.ud[...,0], state.mesh, [0,1,2], C[0], Bl[0], Br[0], second_order_boundary=self._gradient_second_order_boundary)[:]
                grad_ud_y[:] = gradient_with_pbc(state.ud[...,1], state.mesh, [0,1,2], C[1], Bl[1], Br[1], second_order_boundary=self._gradient_second_order_boundary)[:]
                grad_ud_z[:] = gradient_with_pbc(state.ud[...,2], state.mesh, [0,1,2], C[2], Bl[2], Br[2], second_order_boundary=self._gradient_second_order_boundary)[:]
                
                diff_data.set_B_jump_conditions(Bl, Br)

        diff_data.set_gradient_ud(grad_ud_x, grad_ud_y, grad_ud_z)

        self.diff_data = diff_data

    # ----------------------------------------------------
    # PDE RHS setters
    # ----------------------------------------------------
    
    @timedmethod
    def get_fm(self, state):
        # Returns the magnetic part of the force field
        m = state.m
        slice_m = self.slice_m
        
        """ magnetic strain forces """
        mesh = state.mesh
        n = mesh.n
        f_m = torch.zeros((n+(3,3)))

        for i in range(3):
            for j in range(3):
                for term in self._fm_terms[i][j]:
                    f_m[slice_m+(i,j)] += term(state)
        return f_m
    
    def _get_boundary_f(self, t0, s1, s2, hl, hr): 
        # This is actually the midpoint, since the boundary value is outside the node grid
        f_bdr = s2*hl**2. - t0*hr**2. + (hr**2.-hl**2.)*s1
        f_bdr /= hr*hl**2. + hl*hr**2.
        return f_bdr

    #@timedmethod
    def dpd(self, state):
        """ constants """
        n = state.mesh.n

        """ get first derivatives """
        self._update_diff_data(state)

        """ get magnetic stress """
        eps_m = epsilon_m(state)
        sig_m = sigma(state, eps_m)

        """ get stress """
        sig_ii = torch.zeros(n + (3,))
        for i in range(3):
            for sig_term in self._sig_terms[i][i]:
                sig_ii[:,:,:,i] += sig_term(state)

        
        sig_ij = torch.zeros(n + (3,))
        for sig_term in self._sig_terms[1][2]:
            sig_ij[:,:,:,0] += sig_term(state)

        for sig_term in self._sig_terms[0][2]:
            sig_ij[:,:,:,1] += sig_term(state)

        for sig_term in self._sig_terms[0][1]:
            sig_ij[:,:,:,2] += sig_term(state)

        sig_ii -= sig_m[...,:3]
        sig_ij -= sig_m[...,3:]

        """ get bulk forces """
        f_ij = torch.zeros(n + (3,3))

        for i in range(3):
            for j in range(3):
                for term in self._f_terms[i][j]:
                    f_ij[:,:,:,i,j] += term(state)

        """ get expanded dx """
        dx1_exp =  state.mesh.dx_tensor[0].unsqueeze(1).unsqueeze(2).expand_as(state.ud[...,0])
        dx2_exp =  state.mesh.dx_tensor[1].unsqueeze(0).unsqueeze(2).expand_as(state.ud[...,0])
        dx3_exp =  state.mesh.dx_tensor[2].unsqueeze(0).unsqueeze(0).expand_as(state.ud[...,0])
        dx_exp = dx1_exp, dx2_exp, dx3_exp

        """ set boundary conditions """
        ft = torch.zeros(state.mesh.n + (3,3))
        ft_weigth = torch.zeros(state.mesh.n + (3,3), dtype=int)
        for t_bc in self._neumann_bcs:
            t_dim = t_bc.plane.dim
            t_trans_dim1 = t_bc.plane.trans_dim1
            t_trans_dim2 = t_bc.plane.trans_dim2
            t_sign = t_bc.plane.sign
            t_val = t_bc.condition(state)
            t_slc = t_bc.plane.slices

            if self._bc_second_order and n[t_dim] > 1:
                if self._gradient_second_order_boundary:
                    hl = 0.5*dx_exp[t_dim][t_slc] # distance to from s1 to boundary
                    hr = hl + 0.5*(torch.roll(dx_exp[t_dim], t_sign, dims=t_dim)[t_slc]) # distance between s1 and s2

                    # t bc: main diagonal
                    t0 = t_val[...,t_dim]
                    s1 = sig_ii[t_slc+(t_dim,)] # sig value on node
                    s2 = torch.roll(sig_ii[...,t_dim], t_sign, dims=t_dim)[t_slc] # sig value on next node            

                    g_i = self._get_boundary_f(t0, s1, s2, hl, hr)
                    g_i *= -t_sign
                    
                    # t bc: remaining out of plane derivatives
                    t0 = t_val[...,t_trans_dim1]
                    s1 = sig_ij[t_slc+(t_trans_dim2,)] # Note: sig_ij is counted "inverse", as yz, xz, xy 
                    s2 = torch.roll(sig_ij[...,t_trans_dim2], t_sign, dims=t_dim)[t_slc] # Note: sig_ij is counted "inverse", as yz, xz, xy 

                    g_j = self._get_boundary_f(t0, s1, s2, hl, hr)
                    g_j *= -t_sign

                    t0 = t_val[...,t_trans_dim2]
                    s1 = sig_ij[t_slc+(t_trans_dim1,)] # Note: sig_ij is counted "inverse", as yz, xz, xy 
                    s2 = torch.roll(sig_ij[...,t_trans_dim1], t_sign, dims=t_dim)[t_slc] # Note: sig_ij is counted "inverse", as yz, xz, xy 

                    g_k = self._get_boundary_f(t0, s1, s2, hl, hr)
                    g_k *= -t_sign

                else:
                    h = dx_exp[t_dim][t_slc] + 0.5*(torch.roll(dx_exp[t_dim], t_sign, dims=t_dim)[t_slc])

                    # t bc: main diagonal
                    s_bdr = torch.roll(sig_ii[...,t_dim], t_sign, dims=t_dim)[t_slc]
                    g_i = t_sign*(t_val[:,:,t_dim] - s_bdr)/h # compute value
                    
                    # t bc: remaining out of plane derivatives
                    s_bdr = torch.roll(sig_ij[...,t_trans_dim2], t_sign, dims=t_dim)[t_slc] # Note: sig_ij is counted "inverse", as yz, xz, xy 
                    g_j = t_sign*(t_val[:,:,t_trans_dim1] - s_bdr)/h
                    
                    s_bdr = torch.roll(sig_ij[...,t_trans_dim1], t_sign, dims=t_dim)[t_slc] # Note: sig_ij is counted "inverse", as yz, xz, xy 
                    g_k = t_sign*(t_val[:,:,t_trans_dim2] - s_bdr)/h

            else:
                # Note: When forward and backward difference are used in this scheme for the boundary values of sig,
                # they are better estimates for the value at the cell boundary (Mean Value Theorem). 
                # Thus, h is the full cell width dx here
                h = dx_exp[t_dim][t_slc]

                # t bc: main diagonal
                s_bdr = sig_ii[t_slc+(t_dim,)] # sig value on node
                g_i = t_sign*(t_val[:,:,t_dim] - s_bdr)/h # compute value
                
                # t bc: remaining out of plane derivatives
                s_bdr = sig_ij[t_slc+(t_trans_dim2,)] # Note: sig_ij is counted "inverse", as yz, xz, xy 
                g_j = t_sign*(t_val[:,:,t_trans_dim1] - s_bdr)/h
                
                s_bdr = sig_ij[t_slc+(t_trans_dim1,)] # Note: sig_ij is counted "inverse", as yz, xz, xy 
                g_k = t_sign*(t_val[:,:,t_trans_dim2] - s_bdr)/h


            ft[t_slc+(t_dim, t_dim)] += g_i 
            ft[t_slc+(t_trans_dim1, t_dim)] += g_j 
            ft[t_slc+(t_trans_dim2, t_dim)] += g_k

            ft_weigth[t_slc+(t_dim, t_dim)] += 1
            ft_weigth[t_slc+(t_trans_dim1, t_dim)] += 1 
            ft_weigth[t_slc+(t_trans_dim2, t_dim)] += 1
        
        """ get forces due to magnetic strain """
        f_ij -= self.get_fm(state)

        """ set boundary conditions """
        bc_mask = ft_weigth > 0 
        f_ij[bc_mask] = ft[bc_mask] / ft_weigth[bc_mask]

        """ add up all force contributions """
        self.f_ij = f_ij
        f_el = f_ij.sum(dim=-1)

        eta = state.material["eta"]
        return (f_el - eta*state.pd) * self._get_mask_elastic(state)

    @timedmethod
    def dud(self, state):
        rho = state.material["rho"]

        return (state.pd / rho) * self._get_mask_elastic(state) 

    def dv(self, t, v, state, alpha = None):
        
        self._set_solution_variables(state, v)

        # LLG 
        dm = self.dm(t, state.m, state, alpha) # sets t and m before dm is calculated

        # LE
        dud = self.dud(state)
        dpd = self.dpd(state)

        # composite variable
        n = state.mesh.n
        dv = torch.zeros((n[0],n[1],n[2],9))

        dv[:,:,:,:3] = dm
        dv[:,:,:,3:6] = dud
        dv[:,:,:,6:] = dpd

        return dv

    def E(self, state):
        return sum([term.E(state) for term in self._terms])
    
    def U_el(self, state):
        eps = epsilon(state, second_order_boundary=self._gradient_second_order_boundary)
        eps_el = eps - epsilon_m(state)
        sig_el = sigma(state, eps_el)

        zeta = 0.5*eps_el*sig_el

        E_el =  zeta * state.mesh.cell_volumes
        return E_el.sum()
    
    def U(self, state):
        eps = epsilon(state, second_order_boundary=self._gradient_second_order_boundary)
        sig = sigma(state, eps)

        zeta = 0.5*eps*sig

        E_el =  zeta * state.mesh.cell_volumes
        return E_el.sum()

    def T_el(self, state):
        T = 0.5 * state.pd**2. / state.material["rho"]
        return (T*state.mesh.cell_volumes*self._get_mask_elastic(state)).sum()
    
    def _update_neumann_bcs(self, state):
        self._neumann_bcs = []

        def _t_add(state, t0, t1_cond, trans_slice1, trans_slice2):
            t0[trans_slice1, trans_slice2] += t1_cond(state)
            return t0 
        
        def _apply_t_additions(state, t0, addition_list):
            t_out = torch.clone(t0)
            for addition in addition_list:
                t_out = addition(state, t_out)
            return t_out
        
        t_additions = [[[],[]],[[],[]],[[],[]]] # Note: 3*[[[],[]]] due to call by reference not possible
        
        # check for Neumann bcs set by the user
        if hasattr(state, "bcs"):
            if "t" in state.bcs:
                for bc in state.bcs["t"]:
                    # check if there is a conflict with natural boundary conditions
                    pl = bc.plane
                    i_where = None
                    if pl.pos == 0:
                        assert pl.sign == -1
                        i_where = 0
                    elif pl.pos == -1 or pl.pos == state.mesh.n[pl.dim]-1:
                        assert pl.sign == 1
                        i_where = -1

                    # CASE: conflict, mark override
                    if i_where != None:
                        assert state.mesh.pbc[pl.dim] == 0
                        t_additions[pl.dim][i_where].append(lambda state, t0, bc=bc, pl=pl : _t_add(state, t0, bc.condition, pl.trans_slice1, pl.trans_slice2))
                    # CASE: no conflict, use bc as is
                    else:
                        self._neumann_bcs.append(bc)

        # set natural boundary conditions where no pbc are set
        for i in range(3):
            if state.mesh.pbc[i] == 0:
                plane_m = Plane(i, 0, -1)
                plane_p = Plane(i, -1, 1)

                # remove the dimension in which the plane lies from n
                n_bc = list(state.mesh.n)
                n_bc.pop(i)
                n_bc = tuple(n_bc)

                # set homogenous Neumann boundary conditions
                tm = torch.zeros(n_bc + (3,))
                tp = torch.zeros(n_bc + (3,))

                # check for overrides
                if (len(t_additions[i][0]) > 0):
                    cond_m = lambda state, t0=tm, additions=t_additions[i][0]: _apply_t_additions(state, t0, additions)
                else:
                    cond_m = lambda state, t0=tm : t0 

                if (len(t_additions[i][1]) > 0):
                    cond_p = lambda state, t0=tp, additions=t_additions[i][1] : _apply_t_additions(state, t0, additions)
                else:
                    cond_p = lambda state, t0=tp : t0

                # set boundaries
                bc_plane_m = PlaneBC(plane_m, cond_m)
                bc_plane_p = PlaneBC(plane_p, cond_p)
                
                self._neumann_bcs.append(bc_plane_m)
                self._neumann_bcs.append(bc_plane_p)

    @timedmethod
    def step(self, state, dt, rtol = 1e-5, atol = None, atol_m = 1e-5, atol_ud = 1e-15, atol_pd = 1e-2, **kwargs):
        self._update_neumann_bcs(state)
        v_in = self._get_solution_variables(state)

        if not isinstance(atol, torch.Tensor):
            atol = tuple(3*[atol_m] + 3*[atol_ud] + 3*[atol_pd])
            atol = torch.tensor(atol)[None,None,None,:]

        state.t, v_out = self._solver.step(state.t, v_in, dt, state=state, rtol=rtol, atol=atol, **kwargs)
        self._set_solution_variables(state, v_out)
        normalize(state.m)
        logging.info_blue("[LLG + LE] step: dt= %g  t=%g" % (dt, state.t))

    @timedmethod
    def relax(self, state, maxiter = 500, dm_tol = 1e2, dud_tol = 1e-4, dt = 1e-11, rtol=1e-5, atol_m = 1e-5, atol_ud = 1e-15, atol_pd = 1e-2):
        t0 = state.t

        eta_in = state.material["eta"]
        state.material["eta"] = state.Constant(1e10)

        atol = tuple(3*[atol_m] + 3*[atol_ud] + 3*[atol_pd])
        atol = torch.tensor(atol)[None,None,None,:]

        self._update_neumann_bcs(state)

        for i in range(maxiter):
            # step
            v_in = self._get_solution_variables(state)
            state.t, v_out = self._solver.step(state.t, v_in, dt, state=state, rtol=rtol, atol=atol, alpha = 1.0)
            self._set_solution_variables(state, v_out)

            # determine rate of change of m
            dm = self.dm(state.t, state.m, state=state, alpha = 1.0)
            dm = dm[self.slice_m + (slice(None,3),)].abs().max() / constants.gamma # use same scaling as within minimizer

            # determine rate of change of ud
            dud = state.pd / state.material["rho"]
            dud = dud[self._get_mask_elastic(state)].nan_to_num(posinf=0, neginf=0).abs().max()

            logging.info_blue("[LLG+LE] relax: i=%d t=%g |dm|=%g, |dud|=%g" % (i, state.t-t0, dm, dud))
            if dm < dm_tol and dud < dud_tol:
                logging.info_green("[LLG+LE] relax: Successfully converged (iter=%d, dm_tol = %g, dud_tol = %g)" % (i, dm_tol, dud_tol))
                state.t = t0
                state.material["eta"] = eta_in
                state.pd = state.Constant((0.,0.,0.))
                return True

        logging.warning("[LLG+LE] relax: Terminated after maxiter = %d (dm = %g, dm_tol = %g, dud = %g, dud_tol = %g)" % (maxiter, dm, dm_tol,  dud, dud_tol))
        state.t = t0
        state.material["eta"] = eta_in
        state.pd = state.Constant((0.,0.,0.))
        return False



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
import os
from collections.abc import Iterable
from functools import reduce
from magnumnp.common import logging

__all__ = ["ScalarLogger"]

class ScalarLogger(object):
    def __init__(self, filename, columns, every = 1, fsync_every = 1):
        """
        Simple logger class to log scalar values into a tab separated file.

        *Arguments*
            filename (:class:`str`)
                The name of the log file
            columns ([:class:`str` | :class:`function`])
                The columns to be written to the log file
            every (:class:`int`)
                Write row to log file every nth call
            fsync_every (:class:`int`)
                Call fsync every nth write to empty OS buffer

        *Example*
            .. code-block:: python

                # provide key strings with are available in state
                logger = ScalarLogger('log.dat', ['t','m'])

                # provide func(state) or tuple (name, func(state))
                logger = ScalarLogger('log.dat', [('t[ns]', lambda state: state.t*1e9)])

                # provide predifined functions
                logger = ScalarLogger('log.dat', [demag.h, demag.E])

                # Actually log a row
                state = State(mesh)
                logger << state
        """
        # create directory if not existent
        if not os.path.dirname(filename) == '' and \
             not os.path.exists(os.path.dirname(filename)):
            try:
                os.makedirs(os.path.dirname(filename))
            except OSError as exc: # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise

        self._filename = filename
        self._file     = None
        self._every    = every
        self._fsync_every = fsync_every * every
        self._i        = 0
        self._columns  = columns

    def add_column(self, column):
        if self._file is not None:
            raise RuntimeError("You cannot add columns after first log row has been written.")
        self._columns.append(column)

    def log(self, state):
        self._i += 1
        if ((self._i - 1) % self._every > 0):
            return

        values = []

        for column in self._columns:
            if isinstance(column, str):
                name = column
                raw_value = getattr(state, column)
            elif hasattr(column, '__call__'):
                try:
                    name = column.__self__.__class__.__name__ + "." + column.__name__
                except:
                    name = 'unnamed'
                raw_value = column(state)
            elif isinstance(column, tuple) or isinstance(column, list):
                name = column[0]
                raw_value = column[1](state)
            else:
                raise RuntimeError('Column type not supported.')

            if isinstance(raw_value, torch.Tensor):
                value = state.avg(raw_value).tolist()
            else:
                value = raw_value
            values.append((name, value))
        if self._file is None:
            self._file = open(self._filename, 'w')
            self._write_header(values)

        self._write_row(values)
        if ((self._i - 1) % self._fsync_every == 0):
            os.fsync(self._file.fileno())

    def __lshift__(self, state):
        self.log(state)

    def _write_header(self, columns):
        headings = []

        for column in columns:
            if isinstance(column[1], Iterable):
                if (len(column[1]) == 3):
                    for i in ('x', 'y', 'z'):
                        headings.append(column[0] + '_' + i)
                else:
                    for i in range(len(column[1])):
                        headings.append(column[0] + '_' + str(i))
            else:
                headings.append(column[0])

        format_str = "#" + "    ".join(["%-22s"] * len(headings)) + "\n"
        self._file.write(format_str % tuple(headings))
        self._file.flush()

    def _write_row(self, columns):
        flat_values = reduce(lambda x,y: x+y,
            map(lambda x: tuple(x[1]) if isinstance(x[1], Iterable) else (x[1],), columns))
        format_str = "    ".join(["%+.15e"] * len(flat_values)) + "\n"
        self._file.write(format_str % flat_values)
        self._file.flush()

    def __del__(self):
        if self._file is not None:
            self._file.close()

    def resumable_step(self):
        """
        Returns the last step the logger can resume from, e.g. if the logger
        logs every 10th step and the first (i = 0) step was already logged,
        the result is 10.

        *Returns*
            :class:`int`
                The step number the logger is able to resume from
        """
        if self._file is not None:
            raise RuntimeError("Cannot resume from log file that is already open for writing.")

        i = 0
        with open(self._filename, 'r') as f:
            for i, l in enumerate(f):
                pass
        return i * self._every

    def resume(self, i):
        """
        Try to resume existing log file from log step i. The log file
        is truncated accordingly.

        *Arguments*
            i (:class:`int`)
                The log step to resume from
        """
        number = (self.resumable_step() - i) / self._every

        # from https://superuser.com/questions/127786/efficiently-remove-the-last-two-lines-of-an-extremely-large-text-file
        count = 0
        #with open(self._filename, 'r+b') as f:
        #    f.seek(0, os.SEEK_END)
        #    end = f.tell()
        #    while f.tell() > 0:
        #        f.seek(-1, os.SEEK_CUR)
        #        char = f.read(1)
        #        if char != '\n' and f.tell() == end:
        #            raise RuntimeError("Cannot resume: logfile does not end with a newline.")
        #        if char == '\n':
        #            count += 1
        #        if count == number + 1:
        #            f.truncate()
        #            break
        #        f.seek(-1, os.SEEK_CUR)

        with open(self._filename, 'r+b', buffering=0) as f:
            f.seek(0, os.SEEK_END)
            end = f.tell()
            while f.tell() > 0:
                f.seek(-1, os.SEEK_CUR)
                #print(f.tell())
                char = f.read(1)
                if char != b'\n' and f.tell() == end:
                    raise RuntimeError("Cannot resume: logfile does not end with a newline.")
                    #print ("No change: file does not end with a newline")
                    #exit(1)
                if char == b'\n':
                    count += 1
                if count == number + 1:
                    f.truncate()
                    break
                    #print ("Removed " + str(number) + " lines from end of file")
                    #exit(0)
                f.seek(-1, os.SEEK_CUR)

        self._i = i
        if self._i > 0:
            self._file = open(self._filename, 'a')



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
from magnumnp.common import logging

__all__ = ["Heun"]

class Heun(object):
    def __init__(self, f, dt = 5e-15):
        self._f = f
        self._dt = dt

        logging.info_green("[LLGSolver] using Heun solver (dt = %g)" % dt)

    def _f_wrapper(self, state, t, m, **llg_args):
        t0 = state.t
        m0 = state.m.detach()
        state.t = t
        state.m = m
        f = self._f(state, **llg_args)
        state.t = t0
        state.m = m0
        return f

    def step(self, state, dt, **llg_args):
        f, m, t = self._f_wrapper, state.m, state.t

        state._dt = dt  # update current dt in state used by thermal field class
        k1 = dt * f(state, t,      m,      **llg_args)
        k2 = dt * f(state, t + dt, m + k1, **llg_args)
    
        state.m = m + (k1 + k2) / 2.
        state.t = t + dt
        state._step += 1



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch 
import torch._dynamo 

__all__ = ["DerivTermCompiler",
           "EpsTerm",
           "SigTerm",
           "EpsMTerm",
           "SigMTerm",
           "ForceMComponentTerm",
           "ForceComponentTerm"]

# TODO: replace torch.compile by torch.jit.trace(f, example_tensor)
torch._dynamo.config.cache_size_limit = 1024
torch._dynamo.config.suppress_errors = True

class DerivTermCompiler:
    # this object maps stress and force terms on lambda functions that return the given term as torch.tensor

    def compile(self, llg_with_le_solver, term):
        if isinstance(term, ForceComponentTerm):
            return torch.compile(self._compile_f_term(llg_with_le_solver, term))
        elif isinstance(term, ForceMComponentTerm):
            return torch.compile(self._compile_fm_term(llg_with_le_solver, term))
        elif isinstance(term, SigTerm):
            return torch.compile(self._compile_sig_term(llg_with_le_solver, term))
        else:
            raise SystemError("The derivative compiler does not recognize the type of the term provided to it. The type of the term is " + type(term) + " but should be ForceComponentTerm, ForceMComponentTerm or SigTerm!")

    def _compile_sig_term(self, solver, term):
        return lambda state, _solver=solver, iu=term.i_u, ix=term.i_x, ij=term.ij_C : \
            _solver._weighted_first_derivative(state, iu, ix, ij)
    
    def _compile_f_term(self, solver, term):
        if term.i_x1 == term.i_x2:
            return lambda state, _solver=solver, iu=term.i_u, ix=term.i_x1, ij=term.ij_C : \
                _solver._2nd_derivative(state, iu, ix, ij)     
        else:
            return lambda state, _solver=solver, iu=term.i_u, ix1=term.i_x1, ix2=term.i_x2, ij=term.ij_C : \
                _solver._mixed_derivative(state, iu, ix1, ix2, ij)
        
    def _compile_fm_term(self, solver, term):
        if term.i_m == term.j_m:
            return lambda state, _solver=solver, im=term.i_m, ix=term.i_x, ij=term.ij_C : \
                _solver._main_diag_sigM_derivative(state, im, ix, ij)
        else:
            return lambda state, _solver=solver, im=term.i_m, jm=term.j_m, ix=term.i_x, ij=term.ij_C : \
                _solver._off_diag_sigM_derivative(state, im, jm, ix, ij)

class EpsTerm:
    # information on the strain tensor terms
    def __init__(self, i_u, i_x):
        self._iu = i_u
        self._ix = i_x

    # multiplication with a component of the stiffness tensor yields a term of the stress tensor
    def multiply_Cij(self,i,j):
        return SigTerm(self._iu, self._ix, [i,j])

    # index of the displacement component that is differentiated
    @property
    def i_u(self):
        return self._iu
    
    # corresponding dimension in which the displacement is differentiated
    @property
    def i_x(self):
        return self._ix

class SigTerm:
    # information on the stress tensor terms
    def __init__(self, i_u, i_x, ij_C):
        self._iu = i_u
        self._ix = i_x
        self._ijC = ij_C

    # differentiation of stress terms yields terms of the force field
    def differentiate(self, j_x):
        return ForceComponentTerm(self._iu, self._ix, j_x, self._ijC)
    
    # indices (list with two entries, i and j) of the stiffness matrix that weights the derivative
    @property
    def ij_C(self):
        return self._ijC
    
    # index of the displacement component that is differentiated
    @property
    def i_u(self):
        return self._iu
    
    # corresponding dimension in which the displacement is differentiated
    @property
    def i_x(self):
        return self._ix
    
class EpsMTerm:
    def __init__(self, i_m, j_m):
        self._i_m = i_m 
        self._j_m = j_m 
    
    def multiply_Cij(self,i,j):
        return SigMTerm(self._i_m, self._j_m, [i,j])

    @property
    def i_m(self):
        return self._i_m
    
    @property
    def j_m(self):
        return self._j_m

class SigMTerm:
    def __init__(self, i_m, j_m, ij_C):
        self._i_m = i_m 
        self._j_m = j_m 
        self._ij_C = ij_C

    def differentiate(self, j_x):
        return ForceMComponentTerm(self._i_m, self._j_m, j_x, self._ij_C)

    @property
    def i_m(self):
        return self._i_m
    
    @property
    def j_m(self):
        return self._j_m
    
    @property
    def ij_C(self):
        return self._ij_C
    
class ForceMComponentTerm:
    def __init__(self, i_m, j_m, i_x, ij_C):
        self._i_m = i_m
        self._j_m = j_m
        self._i_x = i_x
        self._ij_C = ij_C

    @property
    def i_m(self):
        return self._i_m
    
    @property
    def j_m(self):
        return self._j_m
    
    # first derivative
    @property
    def i_x(self):
        return self._i_x
    
    @property
    def ij_C(self):
        return self._ij_C

class ForceComponentTerm:
    # information on the force field terms
    def __init__(self, i_u, i_x1, i_x2, ij_C):
        self._iu = i_u
        self._ix1 = i_x1
        self._ix2 = i_x2
        self._ijC = ij_C

    # as in SigTerm
    @property
    def i_u(self):
        return self._iu
    
    # first derivative
    @property
    def i_x1(self):
        return self._ix1
    
    # second derivative
    @property
    def i_x2(self):
        return self._ix2
    
    # as in SigTerm
    @property
    def ij_C(self):
        return self._ijC



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import logging, constants, write_vti, complex_dtype, timedmethod
import os
import torch
import numpy as np
from scipy.sparse.linalg import LinearOperator, aslinearoperator, eigs
from scipy.interpolate import interp2d
from scipy.linalg import eig
from xml.etree import cElementTree
from xml.dom import minidom

__all__ = ["EigenSolver", "EigenResult"]

class EigenSolver(object):
    def __init__(self, state, linear_terms, constant_terms, domain=Ellipsis):
        """
        This class implements solution of the linearized LLG in the frequncy domain [dAquino2009]_.
        The corresponding eigenvalue problem is solved using Scipy (CPU only).

        *Example*
            .. code:: python

                llg = LLGSolver([demag, exchange, external])
                logger = Logger("data", ['t', 'm'])
                while state.t < 1e-9-eps:
                    llg.step(state, 1e-11)
                    logger << state

        *Arguments*
            terms ([:class:`LLGTerm`])
                List of LLG contributions to be considered for energy minimization
            state ([:class:`State`])
                the State object containing the equilibrium magnetization state.m
            linear_terms (list)
                list of linear field terms
            constant_terms (list)
                list of constant field terms
            domain ([:class:`torch.Tensor`])
                integrate without precession term (default: False)
        """
        self._domain = domain
        self._linear_terms = linear_terms
        self._state = state
        self._m0 = state.m
        self._h0 = torch.sum(sum([term.h(state) for term in self._linear_terms + constant_terms])*self._m0, dim=-1, keepdim=True)

        ez = state.Constant([1e-15,0.,1.])
        self._e1 = torch.linalg.cross(ez, self._m0)
        self._e1 = self._e1 / torch.linalg.norm(self._e1, axis=3, keepdim=True)
        self._e0 = -torch.linalg.cross(self._e1, self._m0)
        self._e0 = self._e0 / torch.linalg.norm(self._e0, axis=3, keepdim=True)
        self._vv = state.Constant([0.,0.], dtype=torch.complex128)

        self._it = 0

    def _C(self, m):
        self._state.m = m
        return sum([term.h(self._state) for term in self._linear_terms])

    def _D0(self, vv):
        self._it += 1
        if self._it % 500 == 0:
            logging.info_blue("[Eigensolver] it= %d" % self._it)

        vv = torch.from_numpy(vv).to(device=self._state.device)
        vv = vv.reshape(self._vv[self._domain].shape)
        self._vv[...] = 0.
        self._vv[self._domain] = vv

        # apply R
        vvv = self._vv[:,:,:,(0,)]*self._e0 + self._vv[:,:,:,(1,)]*self._e1

        # calculate A0 v = (C+I H0)*v
        hr = self._C(vvv.real)
        hi = self._C(vvv.imag)
        h = hr + 1j*hi
        h -= self._h0*vvv

        # apply B0
        rrr = -constants.gamma * torch.linalg.cross(self._m0.to(dtype = h.dtype), h)

        # apply R^T (reuse vv tensor)
        self._vv[:,:,:,0] = (rrr*self._e0).sum(dim=-1)
        self._vv[:,:,:,1] = (rrr*self._e1).sum(dim=-1)

        return self._vv[self._domain].reshape(-1).detach().cpu().numpy()

    def B0(self, vv):
        return torch.stack([-1j*vv[...,1,:], 1j*vv[...,0,:]], dim=-2)

    @timedmethod
    def solve(self, k=10, tol=1e-6):
        N = np.prod(self._m0[self._domain].shape[:-1])
        D0 = LinearOperator((2*N,2*N), self._D0, dtype=np.complex128)

        evals, evecs2D = eigs(D0, k = 2*k, which = 'SM', tol = tol, v0 = np.ones(2*N))
        #evals, evecs2D = eigs(D0, k = 2*k, sigma = 0, which = 'LM', tol = tol, v0 = np.ones(2*N))

        evalvecs_sorted = sorted(zip(evals.imag,evecs2D.T), key=lambda x: np.abs(x[0]))
        evals = np.array([x[0] for x in evalvecs_sorted if x[0] > 1000.])
        evecs2D = np.array([x[1] for x in evalvecs_sorted if x[0] > 1000.]).transpose()
        evecs2D = torch.from_numpy(evecs2D).to(device=self._state.device)
        evecs2D = evecs2D.reshape(-1,2,evecs2D.shape[-1])

        omega = torch.tensor(evals)

        # NOTE: our mode normalization (phi_i,phi_j)_l2 = \delta_ij seems to differ from the published version!
        #       thus, re-normalize w_i <phi_i,phi_j>_B0 = \delta_ij
        # NOTE: Using magnetic-domain-averaged scalar product: <f,g> = mean_domain(f_i * g_i)
        norm = omega * (evecs2D.conj() * self.B0(evecs2D)).sum(dim=1).mean(dim=0).real
        evecs2D /= torch.sqrt(norm).reshape(1, 1, -1)

        return EigenResult(omega, evecs2D, self._state, domain = self._domain, m0 = self._m0, e0 = self._e0, e1 = self._e1, D0 = D0)


class EigenResult(object):
    def __init__(self, omega, evecs2D, state, **kwargs):
        self._omega = omega
        self._evecs2D = evecs2D
        self._state = state
        self.__dict__.update(kwargs)

    def store(self, filename):
        torch.save({"m0":self.m0, "omega":self._omega, "evecs2D":self._evecs2D, "domain":self.domain}, filename)
        logging.info_green("[Eigensolver] Stored %d eigenvalues to '%s'" % (len(self._omega), filename))

    @staticmethod
    def load(state, filename):
        stored = torch.load(filename, map_location=state.device)
        m0, omega, evecs2D = stored['m0'], stored['omega'], stored['evecs2D']
        domain = stored['domain']
        state.m = m0

        ez = state.Constant([1e-15,0.,1.])
        e1 = torch.linalg.cross(ez, m0)
        e1 = e1 / torch.linalg.norm(e1, axis=3, keepdim=True)
        e0 = -torch.linalg.cross(e1, m0)
        e0 = e0 / torch.linalg.norm(e0, axis=3, keepdim=True)

        logging.info_green("[Eigensolver] Loaded %d eigenvalues from '%s'" % (len(omega), filename))
        return EigenResult(omega, evecs2D, state, domain = domain, m0 = m0, e0 = e0, e1 = e1)

    @property
    def omega(self):
        return self._omega

    @property
    def freq(self):
        return self._omega/2./torch.pi

    def evecs(self, N = slice(None)):
        full = torch.zeros(self.m0.shape[:3] + self._evecs2D.shape[1:], dtype=self._evecs2D.dtype, device=self._evecs2D.device)
        full[self.domain] = self._evecs2D.reshape(full[self.domain].shape)
        vvv = full[:,:,:,(0,),:]*self.e0[:,:,:,:,None] + full[:,:,:,(1,),:]*self.e1[:,:,:,:,None]
        return vvv[...,N]

    def save_evecs3D(self, filename, which = "abs", N = slice(None)):
        if which == "abs":
            op = torch.abs
        elif which == "real":
            op = torch.real
        elif which == "imag":
            op = torch.imag
        else:
            op = which

        freq = self.freq
        evecs = [op(v.squeeze(-1)) for v in self.evecs(N = N).split(1,dim=-1)]
        xmlroot = cElementTree.Element("VTKFile", type="Collection", version="0.1", byte_order="LittleEndian")
        cElementTree.SubElement(xmlroot, "Collection")
        for i, vvv in enumerate(evecs):
            filename_vti = "%s_%04d.vti" % (os.path.splitext(filename)[0], i)
            write_vti(vvv, filename_vti, self._state)
            cElementTree.SubElement(xmlroot[0], "DataSet", timestep=str(self.freq[i].cpu().numpy()), file=os.path.basename(filename_vti))

        with open(filename, 'w') as fd:
            fd.write(minidom.parseString(" ".join(cElementTree.tostring(xmlroot).decode().replace("\n","").split()).replace("> <", "><")).toprettyxml(indent="  "))

    @property
    def domega(self):
        """Compute peak broadening due to damping domega_k = alpha * omega_k^2 * ||phi_k||^2 """
        alpha = self._state.material["alpha"][self.domain].flatten(end_dim=-2).unsqueeze(-1)
        return self._omega**2 * (alpha * (self._evecs2D.conj()*self._evecs2D).real).sum(dim=1).mean(dim=0)

    def _B0(self, vv):
        return torch.stack([-1j * vv[..., 1, :], 1j * vv[..., 0, :]], dim=-2)


    def projection(self, omega, delta_m):
        """Compute volume-averaged PSD using simplified Lorentzian formula.

        Parameters
        ----------
        delta_m : torch.Tensor
            Real-valued deviation field used to obtain modal amplitudes.
        omega : array_like
            Angular frequency axis (rad/s) at which to evaluate the spectrum.

        Returns
        -------
        torch.Tensor
            Volume-averaged PSD evaluated on ``omega``.
        """

        w = torch.tensor(omega)
        w_k = self.omega.unsqueeze(-1)
        dw_k = self.domega.unsqueeze(-1)

        m2d = self._state.Constant([0., 0.])
        m2d[:,:,:,0] = (delta_m * self.e0).sum(axis=-1)
        m2d[:,:,:,1] = (delta_m * self.e1).sum(axis=-1)
        m2d = m2d[self.domain].flatten(end_dim=-2).unsqueeze(-1)
        a_k = (self._omega * (self._evecs2D.conj() * self._B0(m2d)).sum(dim=1).mean(dim=0)).unsqueeze(-1)

        Ms = self._state.material["Ms"][self.domain].flatten(end_dim=-2)
        phi2 = (Ms**2 * (self._evecs2D.conj()*self._evecs2D).real.sum(dim=1)).mean(dim=0).unsqueeze(-1)

        lorentz_pos = 1.0 / ((w - w_k)**2 + dw_k**2)
        lorentz_neg = 1.0 / ((w + w_k)**2 + dw_k**2)
        lorentz = torch.abs(a_k)**2 * w_k**2 * phi2 * (lorentz_pos + lorentz_neg)
        return lorentz.sum(axis=0) / 2 # P(ω) = ⟨Ms² |δm̂|²⟩/2  [A²/m²]


    def spectrum(self, omega, h_excite):
        """Compute volume-averaged absorbed power using Eq. (39) of d'Aquino & Hertel (JAP 133, 033902 (2023)).

        Parameters
        ----------
        omega : array_like
            Angular frequencies (rad/s) at which to evaluate the absorbed power.
        h_excite : torch.Tensor
            Real-valued excitation field dh^ac(x) (same spatial shape as state.m).

        Returns
        -------
        torch.Tensor
            Volume-averaged absorbed power P_abs(omega)
        """

        # calculate h_k = phi_k^H * R^T * P_m0 * h_excite
        h2d = self._state.Constant([0.,0.])
        h2d[:,:,:,0] = (h_excite*self.e0).sum(axis=-1)
        h2d[:,:,:,1] = (h_excite*self.e1).sum(axis=-1)
        h2d = h2d[self.domain].flatten(end_dim=-2).unsqueeze(-1)
        h_k = (self._evecs2D.conj() * h2d).sum(dim=1).mean(dim=0).unsqueeze(-1)

        # calculate coefficiencs a_k = (omega_k/(omega_k-omega+i*domega_k) phi_k^H * R^T * P_m0 * h_excite)
        w = torch.tensor(omega)
        w_k = self.omega.unsqueeze(-1)
        w_k_prime = (self.omega + 1j * self.domega).unsqueeze(-1)
        a_k_pos = constants.gamma * h_k * w_k_prime / (w_k_prime - w)
        a_k_neg = constants.gamma * h_k * w_k_prime / (w_k_prime + w)

        # phi2 = Ms² ||phi_k||^2: sum over 2 components, Ms²-weighted mean over space
        Ms = self._state.material["Ms"][self.domain].flatten(end_dim=-2)
        phi2 = (Ms**2 * (self._evecs2D.conj()*self._evecs2D).real.sum(dim=1)).mean(dim=0).unsqueeze(-1)
        p = 0.5 * phi2 * ((a_k_pos.conj()*a_k_pos).real + (a_k_neg.conj()*a_k_neg).real)
        return p.sum(axis=0) # P(ω) = ⟨Ms² |δm̂|²⟩/2  [A²/m²]


    def absorption(self, omega, h_excite):
        """Compute volume-averaged absorbed power density using Eq. (40) of d'Aquino & Hertel (JAP 133, 033902 (2023)).

        Parameters
        ----------
        omega : array_like
            Angular frequencies (rad/s) at which to evaluate the absorbed power.
        h_excite : torch.Tensor
            Real-valued excitation field dh^ac(x) (same spatial shape as state.m).

        Returns
        -------
        torch.Tensor
            Volume-averaged absorbed power density P_abs(omega) [W/m³]
        """

        # calculate h_k = phi_k^H * R^T * P_m0 * h_excite
        h2d = self._state.Constant([0., 0.])
        h2d[:,:,:,0] = (h_excite*self.e0).sum(axis=-1)
        h2d[:,:,:,1] = (h_excite*self.e1).sum(axis=-1)
        h2d = h2d[self.domain].flatten(end_dim=-2).unsqueeze(-1)
        h_k = (self._evecs2D.conj() * h2d).sum(dim=1).mean(dim=0).unsqueeze(-1)

        # Ms-weighted projection: g_k = mean_domain{ Ms(x) * phi_k^H(x) * h2d(x) }
        Ms = self._state.material["Ms"][self.domain].flatten(end_dim=-2).unsqueeze(-1)
        g_k = (Ms * self._evecs2D.conj() * h2d).sum(dim=1).mean(dim=0).unsqueeze(-1)

        # calculate Pabs = mu_0/2 * Re{i*omega * sum_k g_k^* a_k}
        w = torch.tensor(omega)
        w_k = self.omega.unsqueeze(-1)
        w_k_prime = (self.omega + 1j * self.domega).unsqueeze(-1)
        a_k = constants.gamma * h_k * w_k / (w_k_prime - w)

        Pabs = 0.5 * constants.mu_0 * (1j * w * g_k.conj() * a_k)

        return Pabs.sum(axis=0).squeeze(0).real


    def dispersion(self, points, dx, num_omega=1000):
        vvv = self.evecs()
        m = points(vvv)
        N = m.shape[0]
        kk = 2.*torch.pi * torch.fft.fftshift(torch.fft.fftfreq(N, dx))

        window = torch.hann_window(N, device=m.device)[:,None]
        mfft = torch.abs(torch.fft.fftshift(torch.fft.fft(m * window, dim=0), dim=0))
        mfft = mfft.T

        w = self._omega
        ww = torch.linspace(w.abs().min(), w.abs().max(), num_omega, device=w.device)

        # inline 1D linear interpolation along w
        idx = torch.clamp(torch.searchsorted(w, ww), 1, w.numel() - 1)
        t = (ww - w[idx-1]) / (w[idx] - w[idx-1])
        mm = mfft[idx-1] + (mfft[idx] - mfft[idx-1]) * t.unsqueeze(1)

        return kk, ww, mm
